# Copyright 2024 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utility functions for mods."""


from flwr.client.typing import ClientAppCallable, Mod
from flwr.common import Context, Message


def make_ffn(ffn: ClientAppCallable, mods: list[Mod]) -> ClientAppCallable:
    """."""

    def wrap_ffn(_ffn: ClientAppCallable, _mod: Mod) -> ClientAppCallable:
        def new_ffn(message: Message, context: Context) -> Message:
            return _mod(message, context, _ffn)

        return new_ffn

    for mod in reversed(mods):
        ffn = wrap_ffn(ffn, mod)

    return ffn
